@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2015 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2016 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -40,6 +40,7 @@
 #include <netinet/udp.h>
 #undef __FAVOR_BSD
 
+#include <assert.h>
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -275,6 +276,23 @@ get_option_uint32(struct dhcpcd_ctx *ctx,
 	return 0;
 }
 
+static int
+get_option_uint16(struct dhcpcd_ctx *ctx,
+    uint16_t *i, const struct dhcp_message *dhcp, uint8_t option)
+{
+	const uint8_t *p;
+	size_t len;
+	uint16_t d;
+
+	p = get_option(ctx, dhcp, option, &len);
+	if (!p || len < (ssize_t)sizeof(d))
+		return -1;
+	memcpy(&d, p, sizeof(d));
+	if (i)
+		*i = ntohs(d);
+	return 0;
+}
+
 static int
 get_option_uint8(struct dhcpcd_ctx *ctx,
     uint8_t *i, const struct dhcp_message *dhcp, uint8_t option)
@@ -463,74 +481,6 @@ decode_rfc3361(const uint8_t *data, size_t dl)
 	return sip;
 }
 
-/* Decode an RFC5969 6rd order option into a space
- * separated string. Returns length of string (including
- * terminating zero) or zero on error. */
-ssize_t
-decode_rfc5969(char *out, size_t len, const uint8_t *p, size_t pl)
-{
-	uint8_t ipv4masklen, ipv6prefixlen;
-	uint8_t ipv6prefix[16];
-	uint8_t br[4];
-	int i;
-	ssize_t b, bytes = 0;
-
-	if (pl < 22) {
-		errno = EINVAL;
-		return 0;
-	}
-
-	ipv4masklen = *p++;
-	pl--;
-	ipv6prefixlen = *p++;
-	pl--;
-
-	for (i = 0; i < 16; i++) {
-		ipv6prefix[i] = *p++;
-		pl--;
-	}
-	if (out) {
-		b= snprintf(out, len,
-		    "%d %d "
-		    "%02x%02x:%02x%02x:"
-		    "%02x%02x:%02x%02x:"
-		    "%02x%02x:%02x%02x:"
-		    "%02x%02x:%02x%02x",
-		    ipv4masklen, ipv6prefixlen,
-		    ipv6prefix[0], ipv6prefix[1], ipv6prefix[2], ipv6prefix[3],
-		    ipv6prefix[4], ipv6prefix[5], ipv6prefix[6], ipv6prefix[7],
-		    ipv6prefix[8], ipv6prefix[9], ipv6prefix[10],ipv6prefix[11],
-		    ipv6prefix[12],ipv6prefix[13],ipv6prefix[14], ipv6prefix[15]
-		);
-
-		len -= (size_t)b;
-		out += b;
-		bytes += b;
-	} else {
-		bytes += 16 * 2 + 8 + 2 + 1 + 2;
-	}
-
-	while (pl >= 4) {
-		br[0] = *p++;
-		br[1] = *p++;
-		br[2] = *p++;
-		br[3] = *p++;
-		pl -= 4;
-
-		if (out) {
-			b= snprintf(out, len, " %d.%d.%d.%d",
-			    br[0], br[1], br[2], br[3]);
-			len -= (size_t)b;
-			out += b;
-			bytes += b;
-		} else {
-			bytes += (4 * 4);
-		}
-	}
-
-	return bytes;
-}
-
 static char *
 get_option_string(struct dhcpcd_ctx *ctx,
     const struct dhcp_message *dhcp, uint8_t option)
@@ -583,7 +533,7 @@ route_netmask(uint32_t ip_in)
 /* We need to obey routing options.
  * If we have a CSR then we only use that.
  * Otherwise we add static routes and then routers. */
-struct rt_head *
+static struct rt_head *
 get_option_routes(struct interface *ifp, const struct dhcp_message *dhcp)
 {
 	struct if_options *ifo = ifp->options;
@@ -661,8 +611,9 @@ get_option_routes(struct interface *ifp, const struct dhcp_message *dhcp)
 			{
 				route->gate.s_addr = htonl(INADDR_ANY);
 				route->net.s_addr = htonl(INADDR_BROADCAST);
-			}
-			route->net.s_addr = route_netmask(route->dest.s_addr);
+			} else
+				route->net.s_addr =
+				    route_netmask(route->dest.s_addr);
 			TAILQ_INSERT_TAIL(routes, route, next);
 		}
 	}
@@ -689,6 +640,43 @@ get_option_routes(struct interface *ifp, const struct dhcp_message *dhcp)
 	return routes;
 }
 
+uint16_t
+dhcp_get_mtu(const struct interface *ifp)
+{
+	const struct dhcp_message *dhcp;
+	uint16_t mtu;
+
+	if (ifp->options->mtu)
+		return (uint16_t)ifp->options->mtu;
+	mtu = 0; /* bogus gcc warning */
+	if ((dhcp = D_CSTATE(ifp)->new) == NULL ||
+	    has_option_mask(ifp->options->nomask, DHO_MTU) ||
+	    get_option_uint16(ifp->ctx, &mtu, dhcp, DHO_MTU) == -1)
+		return 0;
+	return mtu;
+}
+
+/* Grab our routers from the DHCP message and apply any MTU value
+ * the message contains */
+struct rt_head *
+dhcp_get_routes(struct interface *ifp)
+{
+	struct rt_head *routes;
+	uint16_t mtu;
+	const struct dhcp_message *dhcp;
+
+	dhcp = D_CSTATE(ifp)->new;
+	routes = get_option_routes(ifp, dhcp);
+	if ((mtu = dhcp_get_mtu(ifp)) != 0) {
+		struct rt *rt;
+
+		TAILQ_FOREACH(rt, routes, next) {
+			rt->mtu = mtu;
+		}
+	}
+	return routes;
+}
+
 #define PUTADDR(_type, _val)						      \
 	{								      \
 		*p++ = _type;						      \
@@ -849,21 +837,27 @@ make_message(struct dhcp_message **message,
 		if (!(ifo->options & DHCPCD_BOOTP)) {
 			int mtu;
 
-			*p++ = DHO_MAXMESSAGESIZE;
-			*p++ = 2;
-			mtu = if_getmtu(ifp->name);
-			if (mtu < MTU_MIN) {
-				if (if_setmtu(ifp->name, MTU_MIN) == 0)
-					sz = MTU_MIN;
+			if ((mtu = if_getmtu(ifp)) == -1)
+				logger(ifp->ctx, LOG_ERR,
+				    "%s: if_getmtu: %m", ifp->name);
+			else if (mtu < MTU_MIN) {
+				if (if_setmtu(ifp, MTU_MIN) == -1)
+					logger(ifp->ctx, LOG_ERR,
+					    "%s: if_setmtu: %m", ifp->name);
+				mtu = MTU_MIN;
 			} else if (mtu > MTU_MAX) {
 				/* Even though our MTU could be greater than
 				 * MTU_MAX (1500) dhcpcd does not presently
 				 * handle DHCP packets any bigger. */
 				mtu = MTU_MAX;
 			}
-			sz = htons((uint16_t)mtu);
-			memcpy(p, &sz, 2);
-			p += 2;
+			if (mtu != -1) {
+				*p++ = DHO_MAXMESSAGESIZE;
+				*p++ = 2;
+				sz = htons((uint16_t)mtu);
+				memcpy(p, &sz, 2);
+				p += 2;
+			}
 		}
 
 		if (ifo->userclass[0]) {
@@ -890,11 +884,7 @@ make_message(struct dhcp_message **message,
 			}
 		}
 
-		if (ifo->hostname[0] == '\0')
-			hostname = get_hostname(hbuf, sizeof(hbuf),
-			    ifo->options & DHCPCD_HOSTNAME_SHORT ? 1 : 0);
-		else
-			hostname = ifo->hostname;
+		hostname = dhcp_get_hostname(hbuf, sizeof(hbuf), ifo);
 
 		/*
 		 * RFC4702 3.1 States that if we send the Client FQDN option
@@ -1137,15 +1127,21 @@ read_lease(struct interface *ifp)
 	uint8_t type;
 	size_t auth_len;
 
-	fd = open(state->leasefile, O_RDONLY);
+	if (state->leasefile[0] == '\0')
+		fd = fileno(stdin);
+	else
+		fd = open(state->leasefile, O_RDONLY);
 	if (fd == -1) {
 		if (errno != ENOENT)
 			logger(ifp->ctx, LOG_ERR, "%s: open `%s': %m",
 			    ifp->name, state->leasefile);
 		return NULL;
 	}
-	logger(ifp->ctx, LOG_DEBUG, "%s: reading lease `%s'",
-	    ifp->name, state->leasefile);
+	if (state->leasefile[0] == '\0')
+		logger(ifp->ctx, LOG_DEBUG, "reading standard input");
+	else
+		logger(ifp->ctx, LOG_DEBUG, "%s: reading lease `%s'",
+		    ifp->name, state->leasefile);
 	dhcp = calloc(1, sizeof(*dhcp));
 	if (dhcp == NULL) {
 		close(fd);
@@ -1158,6 +1154,9 @@ read_lease(struct interface *ifp)
 		return NULL;
 	}
 
+	if (ifp->ctx->options & DHCPCD_DUMPLEASE)
+		return dhcp;
+
 	/* We may have found a BOOTP server */
 	if (get_option_uint8(ifp->ctx, &type, dhcp, DHO_MESSAGETYPE) == -1)
 		type = 0;
@@ -1181,6 +1180,13 @@ read_lease(struct interface *ifp)
 		else
 			logger(ifp->ctx, LOG_DEBUG,
 			    "%s: accepted reconfigure key", ifp->name);
+	} else if ((ifp->options->auth.options & DHCPCD_AUTH_SENDREQUIRE) ==
+	    DHCPCD_AUTH_SENDREQUIRE)
+	{
+		logger(ifp->ctx, LOG_ERR,
+		    "%s: authentication now required", ifp->name);
+		free(dhcp);
+		return NULL;
 	}
 
 	return dhcp;
@@ -1218,12 +1224,13 @@ dhcp_getoption(struct dhcpcd_ctx *ctx,
 		*os = 2; /* code + len */
 		*code = (unsigned int)*od++;
 		*len = (size_t)*od++;
-		if (*len > ol) {
-			errno = EINVAL;
+		if (*len > ol - *os) {
+			errno = ERANGE;
 			return NULL;
 		}
 	}
 
+	*oopt = NULL;
 	for (i = 0, opt = ctx->dhcp_opts; i < ctx->dhcp_opts_len; i++, opt++) {
 		if (opt->option == *code) {
 			*oopt = opt;
@@ -1392,6 +1399,8 @@ get_lease(struct dhcpcd_ctx *ctx,
     struct dhcp_lease *lease, const struct dhcp_message *dhcp)
 {
 
+	assert(dhcp != NULL);
+
 	lease->cookie = dhcp->cookie;
 	/* BOOTP does not set yiaddr for replies when ciaddr is set. */
 	if (dhcp->yiaddr)
@@ -1480,19 +1489,8 @@ dhcp_openudp(struct interface *ifp)
 	char *p;
 #endif
 
-#ifdef SOCK_CLOEXEC
-	if ((s = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, IPPROTO_UDP)) == -1)
+	if ((s = xsocket(PF_INET, SOCK_DGRAM, IPPROTO_UDP, O_CLOEXEC)) == -1)
 		return -1;
-#else
-	if ((s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
-		return -1;
-	if ((n = fcntl(s, F_GETFD, 0)) == -1 ||
-	    fcntl(s, F_SETFD, n | FD_CLOEXEC) == -1)
-	{
-		close(s);
-	        return -1;
-	}
-#endif
 
 	n = 1;
 	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &n, sizeof(n)) == -1)
@@ -1838,23 +1836,39 @@ dhcp_decline(struct interface *ifp)
 }
 
 static void
-dhcp_renew(void *arg)
+dhcp_startrenew(void *arg)
 {
 	struct interface *ifp = arg;
-	struct dhcp_state *state = D_STATE(ifp);
-	struct dhcp_lease *lease = &state->lease;
+	struct dhcp_state *state;
+	struct dhcp_lease *lease;
 
+	if ((state = D_STATE(ifp)) == NULL)
+		return;
+
+	/* Only renew in the bound or renew states */
+	if (state->state != DHS_BOUND &&
+	    state->state != DHS_RENEW)
+		return;
+
+	/* Remove the timeout as the renew may have been forced. */
+	eloop_timeout_delete(ifp->ctx->eloop, dhcp_startrenew, ifp);
+
+	lease = &state->lease;
 	logger(ifp->ctx, LOG_DEBUG, "%s: renewing lease of %s",
 	    ifp->name, inet_ntoa(lease->addr));
-	logger(ifp->ctx, LOG_DEBUG, "%s: rebind in %"PRIu32" seconds,"
-	    " expire in %"PRIu32" seconds",
-	    ifp->name, lease->rebindtime - lease->renewaltime,
-	    lease->leasetime - lease->renewaltime);
 	state->state = DHS_RENEW;
 	state->xid = dhcp_xid(ifp);
+	state->interval = 0;
 	send_renew(ifp);
 }
 
+void
+dhcp_renew(struct interface *ifp)
+{
+
+	dhcp_startrenew(ifp);
+}
+
 static void
 dhcp_arp_announced(struct arp_state *astate)
 {
@@ -1876,6 +1890,7 @@ dhcp_rebind(void *arg)
 	state->state = DHS_REBIND;
 	eloop_timeout_delete(ifp->ctx->eloop, send_renew, ifp);
 	state->lease.server.s_addr = 0;
+	state->interval = 0;
 	ifp->options->options &= ~(DHCPCD_CSR_WARNED |
 	    DHCPCD_ROUTER_HOST_ROUTE_WARNED);
 	send_rebind(ifp);
@@ -1904,11 +1919,29 @@ dhcp_arp_probed(struct arp_state *astate)
 
 	logger(astate->iface->ctx, LOG_DEBUG, "%s: DAD completed for %s",
 	    astate->iface->name, inet_ntoa(astate->addr));
-	dhcp_bind(astate->iface);
-	arp_announce(astate);
+	if (state->state != DHS_INFORM)
+		dhcp_bind(astate->iface);
+#ifndef IN_IFF_TENTATIVE
+	else {
+		struct dhcp_message *oldnew;
+
+		oldnew = state->new;
+		state->new = state->offer;
+		get_lease(astate->iface->ctx, &state->lease, state->new);
+		ipv4_applyaddr(astate->iface);
+		state->new = oldnew;
+	}
+#endif
+
+	/* If we forked, stop here. */
+	if (astate->iface->ctx->options & DHCPCD_FORKED)
+		return;
 
 	/* Stop IPv4LL now we have a working DHCP address */
 	ipv4ll_drop(astate->iface);
+
+	if (ifo->options & DHCPCD_INFORM)
+		dhcp_inform(astate->iface);
 }
 
 static void
@@ -1966,7 +1999,8 @@ dhcp_arp_conflicted(struct arp_state *astate, const struct arp_msg *amsg)
 			astate->failed = astate->addr;
 		arp_report_conflicted(astate, amsg);
 		unlink(state->leasefile);
-		if (!state->lease.frominfo)
+		if (!(ifp->options->options & DHCPCD_STATIC) &&
+		    !state->lease.frominfo)
 			dhcp_decline(ifp);
 #ifdef IN_IFF_DUPLICATED
 		ia = ipv4_iffindaddr(ifp, &astate->addr, NULL);
@@ -1988,10 +2022,14 @@ dhcp_bind(struct interface *ifp)
 	struct dhcp_lease *lease = &state->lease;
 
 	state->reason = NULL;
-	free(state->old);
-	state->old = state->new;
-	state->new = state->offer;
-	state->offer = NULL;
+	/* If we don't have an offer, we are re-binding a lease on preference,
+	 * normally when two interfaces have a lease matching IP addresses. */
+	if (state->offer) {
+		free(state->old);
+		state->old = state->new;
+		state->new = state->offer;
+		state->offer = NULL;
+	}
 	get_lease(ifp->ctx, lease, state->new);
 	if (ifo->options & DHCPCD_STATIC) {
 		logger(ifp->ctx, LOG_INFO, "%s: using static address %s/%d",
@@ -2000,10 +2038,6 @@ dhcp_bind(struct interface *ifp)
 		lease->leasetime = ~0U;
 		state->reason = "STATIC";
 	} else if (ifo->options & DHCPCD_INFORM) {
-		if (ifo->req_addr.s_addr != 0)
-			lease->addr.s_addr = ifo->req_addr.s_addr;
-		else
-			lease->addr.s_addr = state->addr.s_addr;
 		logger(ifp->ctx, LOG_INFO, "%s: received approval for %s",
 		    ifp->name, inet_ntoa(lease->addr));
 		lease->leasetime = ~0U;
@@ -2063,7 +2097,8 @@ dhcp_bind(struct interface *ifp)
 	if (state->reason == NULL) {
 		if (state->old && !(state->added & STATE_FAKE)) {
 			if (state->old->yiaddr == state->new->yiaddr &&
-			    lease->server.s_addr)
+			    lease->server.s_addr &&
+			    state->state != DHS_REBIND)
 				state->reason = "RENEW";
 			else
 				state->reason = "REBIND";
@@ -2076,7 +2111,7 @@ dhcp_bind(struct interface *ifp)
 		lease->renewaltime = lease->rebindtime = lease->leasetime;
 	else {
 		eloop_timeout_add_sec(ifp->ctx->eloop,
-		    (time_t)lease->renewaltime, dhcp_renew, ifp);
+		    (time_t)lease->renewaltime, dhcp_startrenew, ifp);
 		eloop_timeout_add_sec(ifp->ctx->eloop,
 		    (time_t)lease->rebindtime, dhcp_rebind, ifp);
 		eloop_timeout_add_sec(ifp->ctx->eloop,
@@ -2103,6 +2138,9 @@ dhcp_timeout(void *arg)
 	struct dhcp_state *state = D_STATE(ifp);
 
 	dhcp_bind(ifp);
+	/* If we forked, stop here. */
+	if (ifp->ctx->options & DHCPCD_FORKED)
+		return;
 	state->interval = 0;
 	dhcp_discover(ifp);
 }
@@ -2129,6 +2167,71 @@ dhcp_message_new(const struct in_addr *addr, const struct in_addr *mask)
 	return dhcp;
 }
 
+static int
+dhcp_arp_address(struct interface *ifp)
+{
+	struct dhcp_state *state;
+	struct in_addr addr;
+	struct ipv4_addr *ia;
+	struct arp_state *astate;
+
+	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
+
+	state = D_STATE(ifp);
+	state->state = DHS_PROBE;
+	addr.s_addr = state->offer->yiaddr == INADDR_ANY ?
+	    state->offer->ciaddr : state->offer->yiaddr;
+	/* If the interface already has the address configured
+	 * then we can't ARP for duplicate detection. */
+	ia = ipv4_findaddr(ifp->ctx, &addr);
+
+#ifdef IN_IFF_TENTATIVE
+	if (ia == NULL || ia->addr_flags & IN_IFF_NOTUSEABLE) {
+		if ((astate = arp_new(ifp, &addr)) != NULL) {
+			astate->probed_cb = dhcp_arp_probed;
+			astate->conflicted_cb = dhcp_arp_conflicted;
+			astate->announced_cb = dhcp_arp_announced;
+		}
+		if (ia == NULL) {
+			struct dhcp_lease l;
+
+			get_lease(ifp->ctx, &l, state->offer);
+			/* Add the address now, let the kernel handle DAD. */
+			ipv4_addaddr(ifp, &l.addr, &l.net, &l.brd);
+		} else
+			logger(ifp->ctx, LOG_INFO, "%s: waiting for DAD on %s",
+			    ifp->name, inet_ntoa(addr));
+		return 0;
+	}
+#else
+	if (ifp->options->options & DHCPCD_ARP && ia == NULL) {
+		struct dhcp_lease l;
+
+		get_lease(ifp->ctx, &l, state->offer);
+		logger(ifp->ctx, LOG_INFO, "%s: probing address %s/%d",
+		    ifp->name, inet_ntoa(l.addr), inet_ntocidr(l.net));
+		if ((astate = arp_new(ifp, &addr)) != NULL) {
+			astate->probed_cb = dhcp_arp_probed;
+			astate->conflicted_cb = dhcp_arp_conflicted;
+			astate->announced_cb = dhcp_arp_announced;
+			/* We need to handle DAD. */
+			arp_probe(astate);
+		}
+		return 0;
+	}
+#endif
+
+	return 1;
+}
+
+static void
+dhcp_arp_bind(struct interface *ifp)
+{
+
+	if (dhcp_arp_address(ifp) == 1)
+		dhcp_bind(ifp);
+}
+
 static void
 dhcp_static(struct interface *ifp)
 {
@@ -2154,10 +2257,8 @@ dhcp_static(struct interface *ifp)
 
 	state->offer = dhcp_message_new(ia ? &ia->addr : &ifo->req_addr,
 	    ia ? &ia->net : &ifo->req_mask);
-	if (state->offer) {
-		eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
-		dhcp_bind(ifp);
-	}
+	if (state->offer)
+		dhcp_arp_bind(ifp);
 }
 
 void
@@ -2165,41 +2266,52 @@ dhcp_inform(struct interface *ifp)
 {
 	struct dhcp_state *state;
 	struct if_options *ifo;
-	struct ipv4_addr *ap;
+	struct ipv4_addr *ia;
 
 	state = D_STATE(ifp);
 	ifo = ifp->options;
-	if (ifp->ctx->options & DHCPCD_TEST) {
-		state->addr.s_addr = ifo->req_addr.s_addr;
-		state->net.s_addr = ifo->req_mask.s_addr;
-	} else {
-		if (ifo->req_addr.s_addr == INADDR_ANY) {
-			state = D_STATE(ifp);
-			ap = ipv4_iffindaddr(ifp, NULL, NULL);
-			if (ap == NULL) {
-				logger(ifp->ctx, LOG_INFO,
-				    "%s: waiting for 3rd party to "
-				    "configure IP address",
-				    ifp->name);
+
+	state->state = DHS_INFORM;
+	free(state->offer);
+	state->offer = NULL;
+
+	if (ifo->req_addr.s_addr == INADDR_ANY) {
+		ia = ipv4_iffindaddr(ifp, NULL, NULL);
+		if (ia == NULL) {
+			logger(ifp->ctx, LOG_INFO,
+			    "%s: waiting for 3rd party to configure IP address",
+			    ifp->name);
+			if (!(ifp->ctx->options & DHCPCD_TEST)) {
 				state->reason = "3RDPARTY";
 				script_runreason(ifp, state->reason);
+			}
+			return;
+		}
+	} else {
+		ia = ipv4_iffindaddr(ifp, &ifo->req_addr, &ifo->req_mask);
+		if (ia == NULL) {
+			if (ifp->ctx->options & DHCPCD_TEST) {
+				logger(ifp->ctx, LOG_ERR,
+				    "%s: cannot add IP address in test mode",
+				    ifp->name);
 				return;
 			}
-			state->offer =
-			    dhcp_message_new(&ap->addr, &ap->net);
-		} else
-			state->offer =
-			    dhcp_message_new(&ifo->req_addr, &ifo->req_mask);
-		if (state->offer) {
-			ifo->options |= DHCPCD_STATIC;
-			dhcp_bind(ifp);
-			ifo->options &= ~DHCPCD_STATIC;
+			state->offer = dhcp_message_new(&ifo->req_addr,
+			    &ifo->req_mask);
+			if (dhcp_arp_address(ifp) == 0)
+				return;
+			ia = ipv4_iffindaddr(ifp,
+			    &ifo->req_addr, &ifo->req_mask);
+			assert(ia != NULL);
 		}
 	}
 
-	state->state = DHS_INFORM;
-	state->xid = dhcp_xid(ifp);
-	send_inform(ifp);
+	state->offer = dhcp_message_new(&ia->addr, &ia->net);
+	if (state->offer) {
+		state->xid = dhcp_xid(ifp);
+		get_lease(ifp->ctx, &state->lease, state->offer);
+		send_inform(ifp);
+	}
 }
 
 void
@@ -2293,7 +2405,9 @@ dhcp_drop(struct interface *ifp, const char *reason)
 		return;
 	}
 
-	if (ifp->options->options & DHCPCD_RELEASE) {
+	if (ifp->options->options & DHCPCD_RELEASE &&
+	    !(ifp->options->options & DHCPCD_INFORM))
+	{
 		/* Failure to send the release may cause this function to
 		 * re-enter so guard by setting the state. */
 		if (state->state == DHS_RELEASE)
@@ -2322,6 +2436,8 @@ dhcp_drop(struct interface *ifp, const char *reason)
 	dhcp_auth_reset(&state->auth);
 	dhcp_close(ifp);
 
+	free(state->offer);
+	state->offer = NULL;
 	free(state->old);
 	state->old = state->new;
 	state->new = NULL;
@@ -2445,8 +2561,9 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 	unsigned int i;
 	size_t auth_len;
 	char *msg;
+#ifdef IN_IFF_DUPLICATED
 	struct ipv4_addr *ia;
-	struct arp_state *astate;
+#endif
 
 	/* We may have found a BOOTP server */
 	if (get_option_uint8(ifp->ctx, &type, dhcp, DHO_MESSAGETYPE) == -1)
@@ -2478,12 +2595,15 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		else
 			logger(ifp->ctx, LOG_DEBUG,
 			    "%s: accepted reconfigure key", ifp->name);
-	} else if (ifo->auth.options & DHCPCD_AUTH_REQUIRE) {
-		log_dhcp1(LOG_ERR, "no authentication", ifp, dhcp, from, 0);
-		return;
-	} else if (ifo->auth.options & DHCPCD_AUTH_SEND)
+	} else if (ifo->auth.options & DHCPCD_AUTH_SEND) {
+		if (ifo->auth.options & DHCPCD_AUTH_REQUIRE) {
+			log_dhcp1(LOG_ERR, "no authentication",
+			    ifp, dhcp, from, 0);
+			return;
+		}
 		log_dhcp1(LOG_WARNING, "no authentication",
 		    ifp, dhcp, from, 0);
+	}
 
 	/* RFC 3203 */
 	if (type == DHCP_FORCERENEW) {
@@ -2497,7 +2617,8 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		if (auth == NULL) {
 			log_dhcp(LOG_ERR, "unauthenticated Force Renew",
 			    ifp, dhcp, from);
-			return;
+			if (ifo->auth.options & DHCPCD_AUTH_REQUIRE)
+				return;
 		}
 		if (state->state != DHS_BOUND && state->state != DHS_INFORM) {
 			log_dhcp(LOG_DEBUG, "not bound, ignoring Force Renew",
@@ -2507,11 +2628,9 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		log_dhcp(LOG_ERR, "Force Renew from", ifp, dhcp, from);
 		/* The rebind and expire timings are still the same, we just
 		 * enter the renew state early */
-		if (state->state == DHS_BOUND) {
-			eloop_timeout_delete(ifp->ctx->eloop,
-			    dhcp_renew, ifp);
+		if (state->state == DHS_BOUND)
 			dhcp_renew(ifp);
-		} else {
+		else {
 			eloop_timeout_delete(ifp->ctx->eloop,
 			    send_inform, ifp);
 			dhcp_inform(ifp);
@@ -2523,7 +2642,7 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		/* Before we supported FORCERENEW we closed off the raw
 		 * port so we effectively ignored all messages.
 		 * As such we'll not log by default here. */
-		//log_dhcp(LOG_DEBUG, "bound, ignoring", iface, dhcp, from);
+		//log_dhcp(LOG_DEBUG, "bound, ignoring", ifp, dhcp, from);
 		return;
 	}
 
@@ -2535,6 +2654,13 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		    inet_ntoa(*from));
 		return;
 	}
+
+	if (state->state == DHS_PROBE) {
+		/* Ignore any DHCP messages whilst probing a lease to bind. */
+		log_dhcp(LOG_DEBUG, "probing, ignoring", ifp, dhcp, from);
+		return;
+	}
+
 	/* reset the message counter */
 	state->interval = 0;
 
@@ -2671,6 +2797,16 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		    get_option_addr(ifp->ctx,
 		    &lease->server, dhcp, DHO_SERVERID) != 0)
 			lease->server.s_addr = INADDR_ANY;
+
+		/* Test for rapid commit in the OFFER */
+		if (!(ifp->ctx->options & DHCPCD_TEST) &&
+		    has_option_mask(ifo->requestmask, DHO_RAPIDCOMMIT) &&
+		    get_option(ifp->ctx, dhcp, DHO_RAPIDCOMMIT, NULL))
+		{
+			state->state = DHS_REQUEST;
+			goto rapidcommit;
+		}
+
 		log_dhcp(LOG_INFO, "offered", ifp, dhcp, from);
 		free(state->offer);
 		state->offer = dhcp;
@@ -2713,6 +2849,20 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 			return;
 		}
 
+		if (state->state == DHS_DISCOVER) {
+			/* We only allow ACK of rapid commit DISCOVER. */
+			if (has_option_mask(ifo->requestmask,
+			    DHO_RAPIDCOMMIT) &&
+			    get_option(ifp->ctx, dhcp, DHO_RAPIDCOMMIT, NULL))
+				state->state = DHS_REQUEST;
+			else {
+				log_dhcp(LOG_DEBUG, "ignoring ack of",
+				    ifp, dhcp, from);
+				return;
+			}
+		}
+
+rapidcommit:
 		if (!(ifo->options & DHCPCD_INFORM))
 			log_dhcp(LOG_DEBUG, "acknowledged", ifp, dhcp, from);
 		else
@@ -2735,43 +2885,7 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 	lease->frominfo = 0;
 	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
 
-	addr.s_addr = state->offer->yiaddr;
-	/* If the interface already has the address configured
-	 * then we can't ARP for duplicate detection. */
-	ia = ipv4_findaddr(ifp->ctx, &addr);
-
-#ifdef IN_IFF_TENTATIVE
-	if (ia == NULL || ia->addr_flags & IN_IFF_NOTUSEABLE) {
-		if ((astate = arp_new(ifp, &addr)) != NULL) {
-			astate->probed_cb = dhcp_arp_probed;
-			astate->conflicted_cb = dhcp_arp_conflicted;
-			astate->announced_cb = dhcp_arp_announced;
-		}
-		if (ia == NULL) {
-			struct dhcp_lease l;
-
-			get_lease(ifp->ctx, &l, state->offer);
-			/* Add the address now, let the kernel handle DAD. */
-			ipv4_addaddr(ifp, &l.addr, &l.net, &l.brd);
-		}
-		return;
-	}
-#else
-	if (ifo->options & DHCPCD_ARP) {
-		if (ia == NULL) {
-			if ((astate = arp_new(ifp, &addr)) != NULL) {
-				astate->probed_cb = dhcp_arp_probed;
-				astate->conflicted_cb = dhcp_arp_conflicted;
-				astate->announced_cb = dhcp_arp_announced;
-				/* We need to handle DAD. */
-				arp_probe(astate);
-			}
-			return;
-		}
-	}
-#endif
-
-	dhcp_bind(ifp);
+	dhcp_arp_bind(ifp);
 }
 
 static size_t
@@ -3200,7 +3314,7 @@ dhcp_start1(void *arg)
 	}
 	/* We don't want to read the old lease if we NAK an old test */
 	nolease = state->offer && ifp->ctx->options & DHCPCD_TEST;
-	if (!nolease) {
+	if (!nolease && ifo->options & DHCPCD_DHCP) {
 		state->offer = read_lease(ifp);
 		/* Check the saved lease matches the type we want */
 		if (state->offer) {
@@ -3291,13 +3405,8 @@ dhcp_start1(void *arg)
 	}
 
 	if (!(ifo->options & DHCPCD_DHCP)) {
-		if (ifo->options & DHCPCD_IPV4LL) {
-			if (state->offer && state->offer->cookie != 0) {
-				free(state->offer);
-				state->offer = NULL;
-			}
+		if (ifo->options & DHCPCD_IPV4LL)
 			ipv4ll_start(ifp);
-		}
 		return;
 	}
 
@@ -3334,12 +3443,19 @@ dhcp_start(struct interface *ifp)
 	eloop_timeout_add_tv(ifp->ctx->eloop, &tv, dhcp_start1, ifp);
 }
 
+void
+dhcp_abort(struct interface *ifp)
+{
+
+	eloop_timeout_delete(ifp->ctx->eloop, dhcp_start1, ifp);
+}
+
 void
 dhcp_handleifa(int cmd, struct interface *ifp,
 	const struct in_addr *addr,
 	const struct in_addr *net,
 	const struct in_addr *dst,
-	__unused int flags)
+	int flags)
 {
 	struct dhcp_state *state;
 	struct if_options *ifo;
@@ -3365,6 +3481,13 @@ dhcp_handleifa(int cmd, struct interface *ifp,
 	if (cmd != RTM_NEWADDR)
 		return;
 
+#ifdef IN_IFF_NOTUSEABLE
+	if (flags & IN_IFF_NOTUSEABLE)
+		return;
+#else
+	UNUSED(flags);
+#endif
+
 	ifo = ifp->options;
 	if (ifo->options & DHCPCD_INFORM) {
 		if (state->state != DHS_INFORM)
