@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2015 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2016 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -25,9 +25,7 @@
  * SUCH DAMAGE.
  */
 
-const char dhcpcd_copyright[] = "Copyright (c) 2006-2015 Roy Marples";
-
-#define _WITH_DPRINTF /* Stop FreeBSD bitching */
+const char dhcpcd_copyright[] = "Copyright (c) 2006-2016 Roy Marples";
 
 #include <sys/file.h>
 #include <sys/socket.h>
@@ -111,7 +109,7 @@ static void
 usage(void)
 {
 
-printf("usage: "PACKAGE"\t[-46ABbDdEGgHJKkLnpqTVw]\n"
+printf("usage: "PACKAGE"\t[-46ABbDdEGgHJKkLnPpqTVw]\n"
 	"\t\t[-C, --nohook hook] [-c, --script script]\n"
 	"\t\t[-e, --env value] [-F, --fqdn FQDN] [-f, --config file]\n"
 	"\t\t[-h, --hostname hostname] [-I, --clientid clientid]\n"
@@ -228,6 +226,9 @@ dhcpcd_ifafwaiting(const struct interface *ifp)
 {
 	unsigned long long opts;
 
+	if (!ifp->active)
+		return AF_MAX;
+
 	opts = ifp->options->options;
 	if (opts & DHCPCD_WAITIP4 && !ipv4_hasaddr(ifp))
 		return AF_INET;
@@ -317,33 +318,48 @@ dhcpcd_daemonise(struct dhcpcd_ctx *ctx)
 			return 0;
 	}
 
+	if (ctx->options & DHCPCD_ONESHOT) {
+		logger(ctx, LOG_INFO, "exiting due to oneshot");
+		eloop_exit(ctx->eloop, EXIT_SUCCESS);
+		return 0;
+	}
+
 	eloop_timeout_delete(ctx->eloop, handle_exit_timeout, ctx);
 	if (ctx->options & DHCPCD_DAEMONISED ||
 	    !(ctx->options & DHCPCD_DAEMONISE))
 		return 0;
+	logger(ctx, LOG_DEBUG, "forking to background");
+
 	/* Setup a signal pipe so parent knows when to exit. */
 	if (pipe(sidpipe) == -1) {
 		logger(ctx, LOG_ERR, "pipe: %m");
 		return 0;
 	}
-	logger(ctx, LOG_DEBUG, "forking to background");
+
+	/* Store the pid and routing message seq number so we can identify
+	 * the last message successfully sent to the kernel.
+	 * This allows us to ignore all messages we sent after forking
+	 * and detaching. */
+	ctx->ppid = getpid();
+	ctx->pseq = ctx->sseq;
+
 	switch (pid = fork()) {
 	case -1:
 		logger(ctx, LOG_ERR, "fork: %m");
 		return 0;
 	case 0:
 		setsid();
+		/* Notify parent it's safe to exit as we've detached. */
+		close(sidpipe[0]);
+		if (write(sidpipe[1], &buf, 1) == -1)
+			logger(ctx, LOG_ERR, "failed to notify parent: %m");
+		close(sidpipe[1]);
 		/* Some polling methods don't survive after forking,
 		 * so ensure we can requeue all our events. */
 		if (eloop_requeue(ctx->eloop) == -1) {
 			logger(ctx, LOG_ERR, "eloop_requeue: %m");
 			eloop_exit(ctx->eloop, EXIT_FAILURE);
 		}
-		/* Notify parent it's safe to exit as we've detached. */
-		close(sidpipe[0]);
-		if (write(sidpipe[1], &buf, 1) == -1)
-			logger(ctx, LOG_ERR, "failed to notify parent: %m");
-		close(sidpipe[1]);
 		if ((fd = open(_PATH_DEVNULL, O_RDWR, 0)) != -1) {
 			dup2(fd, STDIN_FILENO);
 			dup2(fd, STDOUT_FILENO);
@@ -392,6 +408,9 @@ stop_interface(struct interface *ifp)
 	struct dhcpcd_ctx *ctx;
 
 	ctx = ifp->ctx;
+	if (!ifp->active)
+		goto stop;
+
 	logger(ctx, LOG_INFO, "%s: removing interface", ifp->name);
 	ifp->options->options |= DHCPCD_STOPPING;
 
@@ -404,10 +423,10 @@ stop_interface(struct interface *ifp)
 	/* Delete all timeouts for the interfaces */
 	eloop_q_timeout_delete(ctx->eloop, 0, NULL, ifp);
 
-	/* Remove the interface from our list */
-	TAILQ_REMOVE(ifp->ctx->ifaces, ifp, next);
-	if_free(ifp);
+	/* De-activate the interface */
+	ifp->active = 0;
 
+stop:
 	if (!(ctx->options & (DHCPCD_MASTER | DHCPCD_TEST)))
 		eloop_exit(ctx->eloop, EXIT_FAILURE);
 }
@@ -453,7 +472,7 @@ configure_interface1(struct interface *ifp)
 		ifo->options |= DHCPCD_IPV6RA_OWN;
 
 	/* We want to disable kernel interface RA as early as possible. */
-	if (ifo->options & DHCPCD_IPV6RS &&
+	if (ifo->options & DHCPCD_IPV6 &&
 	    !(ifp->ctx->options & DHCPCD_DUMPLEASE))
 	{
 		/* If not doing any DHCP, disable the RDNSS requirement. */
@@ -651,6 +670,41 @@ dhcpcd_pollup(void *arg)
 	dhcpcd_handlecarrier(ifp->ctx, carrier, ifp->flags, ifp->name);
 }
 
+static void
+dhcpcd_initstate1(struct interface *ifp, int argc, char **argv,
+    unsigned long long options)
+{
+	struct if_options *ifo;
+
+	configure_interface(ifp, argc, argv, options);
+	ifo = ifp->options;
+
+	if (ifo->options & DHCPCD_IPV4 && ipv4_init(ifp->ctx) == -1) {
+		logger(ifp->ctx, LOG_ERR, "ipv4_init: %m");
+		ifo->options &= ~DHCPCD_IPV4;
+	}
+	if (ifo->options & DHCPCD_IPV6 && ipv6_init(ifp->ctx) == NULL) {
+		logger(ifp->ctx, LOG_ERR, "ipv6_init: %m");
+		ifo->options &= ~DHCPCD_IPV6RS;
+	}
+
+	/* Add our link-local address before upping the interface
+	 * so our RFC7217 address beats the hwaddr based one.
+	 * This needs to happen before PREINIT incase a hook script
+	 * inadvertently ups the interface. */
+	if (ifo->options & DHCPCD_IPV6 && ipv6_start(ifp) == -1) {
+		logger(ifp->ctx, LOG_ERR, "%s: ipv6_start: %m", ifp->name);
+		ifo->options &= ~DHCPCD_IPV6;
+	}
+}
+
+static void
+dhcpcd_initstate(struct interface *ifp, unsigned long long options)
+{
+
+	dhcpcd_initstate1(ifp, ifp->ctx->argc, ifp->ctx->argv, options);
+}
+
 void
 dhcpcd_handlecarrier(struct dhcpcd_ctx *ctx, int carrier, unsigned int flags,
     const char *ifname)
@@ -658,7 +712,9 @@ dhcpcd_handlecarrier(struct dhcpcd_ctx *ctx, int carrier, unsigned int flags,
 	struct interface *ifp;
 
 	ifp = if_find(ctx->ifaces, ifname);
-	if (ifp == NULL || !(ifp->options->options & DHCPCD_LINK))
+	if (ifp == NULL ||
+	    ifp->options == NULL || !(ifp->options->options & DHCPCD_LINK) ||
+	    !ifp->active)
 		return;
 
 	switch(carrier) {
@@ -754,7 +810,7 @@ warn_iaid_conflict(struct interface *ifp, uint8_t *iaid)
 	size_t i;
 
 	TAILQ_FOREACH(ifn, ifp->ctx->ifaces, next) {
-		if (ifn == ifp)
+		if (ifn == ifp || !ifn->active)
 			continue;
 		if (memcmp(ifn->options->iaid, iaid,
 		    sizeof(ifn->options->iaid)) == 0)
@@ -852,20 +908,21 @@ dhcpcd_startinterface(void *arg)
 	}
 
 	if (ifo->options & DHCPCD_IPV6) {
-		if (ifo->options & DHCPCD_IPV6RS &&
-		    !(ifo->options & DHCPCD_INFORM))
+		if (ifo->options & DHCPCD_IPV6RS)
 			ipv6nd_startrs(ifp);
 
 		if (ifo->options & DHCPCD_DHCP6)
 			dhcp6_find_delegates(ifp);
 
 		if (!(ifo->options & DHCPCD_IPV6RS) ||
-		    ifo->options & DHCPCD_IA_FORCED)
+		    ifo->options & (DHCPCD_IA_FORCED | DHCPCD_INFORM6))
 		{
 			ssize_t nolease;
 
 			if (ifo->options & DHCPCD_IA_FORCED)
 				nolease = dhcp6_start(ifp, DH6S_INIT);
+			else if (ifo->options & DHCPCD_INFORM6)
+				nolease = dhcp6_start(ifp, DH6S_INFORM);
 			else {
 				nolease = 0;
 				/* Enabling the below doesn't really make
@@ -924,55 +981,6 @@ dhcpcd_prestartinterface(void *arg)
 	dhcpcd_startinterface(ifp);
 }
 
-static void
-handle_link(void *arg)
-{
-	struct dhcpcd_ctx *ctx;
-
-	ctx = arg;
-	if (if_managelink(ctx) == -1) {
-		logger(ctx, LOG_ERR, "if_managelink: %m");
-		eloop_event_delete(ctx->eloop, ctx->link_fd);
-		close(ctx->link_fd);
-		ctx->link_fd = -1;
-	}
-}
-
-static void
-dhcpcd_initstate1(struct interface *ifp, int argc, char **argv,
-    unsigned long long options)
-{
-	struct if_options *ifo;
-
-	configure_interface(ifp, argc, argv, options);
-	ifo = ifp->options;
-
-	if (ifo->options & DHCPCD_IPV4 && ipv4_init(ifp->ctx) == -1) {
-		logger(ifp->ctx, LOG_ERR, "ipv4_init: %m");
-		ifo->options &= ~DHCPCD_IPV4;
-	}
-	if (ifo->options & DHCPCD_IPV6 && ipv6_init(ifp->ctx) == NULL) {
-		logger(ifp->ctx, LOG_ERR, "ipv6_init: %m");
-		ifo->options &= ~DHCPCD_IPV6RS;
-	}
-
-	/* Add our link-local address before upping the interface
-	 * so our RFC7217 address beats the hwaddr based one.
-	 * This needs to happen before PREINIT incase a hook script
-	 * inadvertently ups the interface. */
-	if (ifo->options & DHCPCD_IPV6 && ipv6_start(ifp) == -1) {
-		logger(ifp->ctx, LOG_ERR, "%s: ipv6_start: %m", ifp->name);
-		ifo->options &= ~DHCPCD_IPV6;
-	}
-}
-
-void
-dhcpcd_initstate(struct interface *ifp, unsigned long long options)
-{
-
-	dhcpcd_initstate1(ifp, ifp->ctx->argc, ifp->ctx->argv, options);
-}
-
 static void
 run_preinit(struct interface *ifp)
 {
@@ -988,6 +996,32 @@ run_preinit(struct interface *ifp)
 		    ifp->carrier == LINK_UP ? "CARRIER" : "NOCARRIER");
 }
 
+void
+dhcpcd_activateinterface(struct interface *ifp)
+{
+
+	if (!ifp->active) {
+		ifp->active = 1;
+		dhcpcd_initstate(ifp, 0);
+		run_preinit(ifp);
+		dhcpcd_prestartinterface(ifp);
+	}
+}
+
+static void
+handle_link(void *arg)
+{
+	struct dhcpcd_ctx *ctx;
+
+	ctx = arg;
+	if (if_managelink(ctx) == -1) {
+		logger(ctx, LOG_ERR, "if_managelink: %m");
+		eloop_event_delete(ctx->eloop, ctx->link_fd);
+		close(ctx->link_fd);
+		ctx->link_fd = -1;
+	}
+}
+
 int
 dhcpcd_handleinterface(void *arg, int action, const char *ifname)
 {
@@ -1004,21 +1038,17 @@ dhcpcd_handleinterface(void *arg, int action, const char *ifname)
 			errno = ESRCH;
 			return -1;
 		}
-		logger(ctx, LOG_DEBUG, "%s: interface departed", ifp->name);
-		ifp->options->options |= DHCPCD_DEPARTED;
-		stop_interface(ifp);
+		if (ifp->active) {
+			logger(ctx, LOG_DEBUG, "%s: interface departed",
+			    ifp->name);
+			ifp->options->options |= DHCPCD_DEPARTED;
+			stop_interface(ifp);
+		}
+		TAILQ_REMOVE(ctx->ifaces, ifp, next);
+		if_free(ifp);
 		return 0;
 	}
 
-	/* If running off an interface list, check it's in it. */
-	if (ctx->ifc && action != 2) {
-		for (i = 0; i < ctx->ifc; i++)
-			if (strcmp(ctx->ifv[i], ifname) == 0)
-				break;
-		if (i >= ctx->ifc)
-			return 0;
-	}
-
 	i = -1;
 	ifs = if_discover(ctx, -1, UNCONST(argv));
 	if (ifs == NULL) {
@@ -1028,25 +1058,40 @@ dhcpcd_handleinterface(void *arg, int action, const char *ifname)
 	TAILQ_FOREACH_SAFE(ifp, ifs, next, ifn) {
 		if (strcmp(ifp->name, ifname) != 0)
 			continue;
+
+		/* If running off an interface list, check it's in it. */
+		if (ctx->ifc) {
+			for (i = 0; i < ctx->ifc; i++)
+				if (strcmp(ctx->ifv[i], ifname) == 0)
+					break;
+			if (i >= ctx->ifc)
+				ifp->active = 0;
+		}
+
 		i = 0;
 		/* Check if we already have the interface */
 		iff = if_find(ctx->ifaces, ifp->name);
 		if (iff) {
-			logger(ctx, LOG_DEBUG, "%s: interface updated", iff->name);
+			if (iff->active)
+				logger(ctx, LOG_DEBUG, "%s: interface updated",
+				    iff->name);
 			/* The flags and hwaddr could have changed */
 			iff->flags = ifp->flags;
 			iff->hwlen = ifp->hwlen;
 			if (ifp->hwlen != 0)
 				memcpy(iff->hwaddr, ifp->hwaddr, iff->hwlen);
 		} else {
-			logger(ctx, LOG_DEBUG, "%s: interface added", ifp->name);
 			TAILQ_REMOVE(ifs, ifp, next);
 			TAILQ_INSERT_TAIL(ctx->ifaces, ifp, next);
+			if (!ifp->active)
+				continue;
+			logger(ctx, LOG_DEBUG, "%s: interface added",
+			    ifp->name);
 			dhcpcd_initstate(ifp, 0);
 			run_preinit(ifp);
 			iff = ifp;
 		}
-		if (action > 0)
+		if (action > 0 && iff->active)
 			dhcpcd_prestartinterface(iff);
 	}
 
@@ -1121,60 +1166,82 @@ reload_config(struct dhcpcd_ctx *ctx)
 static void
 reconf_reboot(struct dhcpcd_ctx *ctx, int action, int argc, char **argv, int oi)
 {
-	struct if_head *ifs;
-	struct interface *ifn, *ifp;
+	int i;
+	struct interface *ifp;
 
-	ifs = if_discover(ctx, argc - oi, argv + oi);
-	if (ifs == NULL) {
-		logger(ctx, LOG_ERR, "%s: if_discover: %m", __func__);
-		return;
-	}
-
-	while ((ifp = TAILQ_FIRST(ifs))) {
-		TAILQ_REMOVE(ifs, ifp, next);
-		ifn = if_find(ctx->ifaces, ifp->name);
-		if (ifn) {
+	TAILQ_FOREACH(ifp, ctx->ifaces, next) {
+		for (i = oi; i < argc; i++) {
+			if (strcmp(ifp->name, argv[i]) == 0)
+				break;
+		}
+		if (oi != argc && i == argc)
+			continue;
+		if (ifp->active) {
 			if (action)
-				if_reboot(ifn, argc, argv);
+				if_reboot(ifp, argc, argv);
 			else
-				ipv4_applyaddr(ifn);
-			if_free(ifp);
-		} else {
-			TAILQ_INSERT_TAIL(ctx->ifaces, ifp, next);
+				ipv4_applyaddr(ifp);
+		} else if (i != argc) {
+			ifp->active = 1;
 			dhcpcd_initstate1(ifp, argc, argv, 0);
 			run_preinit(ifp);
 			dhcpcd_prestartinterface(ifp);
 		}
 	}
-	free(ifs);
 }
 
 static void
-stop_all_interfaces(struct dhcpcd_ctx *ctx, int do_release)
+stop_all_interfaces(struct dhcpcd_ctx *ctx, unsigned long long opts)
 {
 	struct interface *ifp;
 
 	/* Drop the last interface first */
-	while ((ifp = TAILQ_LAST(ctx->ifaces, if_head)) != NULL) {
-		if (do_release) {
-			ifp->options->options |= DHCPCD_RELEASE;
-			ifp->options->options &= ~DHCPCD_PERSISTENT;
+	TAILQ_FOREACH_REVERSE(ifp, ctx->ifaces, if_head, next) {
+		if (ifp->options) {
+			ifp->options->options |= opts;
+			if (ifp->options->options & DHCPCD_RELEASE)
+				ifp->options->options &= ~DHCPCD_PERSISTENT;
+			ifp->options->options |= DHCPCD_EXITING;
 		}
-		ifp->options->options |= DHCPCD_EXITING;
 		stop_interface(ifp);
 	}
 }
 
+static void
+dhcpcd_ifrenew(struct interface *ifp)
+{
+
+#define DHCPCD_RARENEW (DHCPCD_IPV6 | DHCPCD_IPV6RS)
+	if (ifp->options->options & DHCPCD_LINK &&
+	    ifp->carrier != LINK_DOWN)
+	{
+		dhcp_renew(ifp);
+		if ((ifp->options->options & DHCPCD_RARENEW) == DHCPCD_RARENEW)
+			ipv6nd_startrs(ifp);
+		dhcp6_renew(ifp);
+	}
+}
+
+static void
+dhcpcd_renew(struct dhcpcd_ctx *ctx)
+{
+	struct interface *ifp;
+
+	TAILQ_FOREACH(ifp, ctx->ifaces, next) {
+		dhcpcd_ifrenew(ifp);
+	}
+}
+
 #ifdef USE_SIGNALS
 #define sigmsg "received %s, %s"
 static void
 signal_cb(int sig, void *arg)
 {
 	struct dhcpcd_ctx *ctx = arg;
-	struct interface *ifp;
-	int do_release, exit_code;
+	unsigned long long opts;
+	int exit_code;
 
-	do_release = 0;
+	opts = 0;
 	exit_code = EXIT_FAILURE;
 	switch (sig) {
 	case SIGINT:
@@ -1186,7 +1253,7 @@ signal_cb(int sig, void *arg)
 		break;
 	case SIGALRM:
 		logger(ctx, LOG_INFO, sigmsg, "SIGALRM", "releasing");
-		do_release = 1;
+		opts |= DHCPCD_RELEASE;
 		exit_code = EXIT_SUCCESS;
 		break;
 	case SIGHUP:
@@ -1198,10 +1265,8 @@ signal_cb(int sig, void *arg)
 		    ctx->argc - ctx->ifc);
 		return;
 	case SIGUSR1:
-		logger(ctx, LOG_INFO, sigmsg, "SIGUSR1", "reconfiguring");
-		TAILQ_FOREACH(ifp, ctx->ifaces, next) {
-			ipv4_applyaddr(ifp);
-		}
+		logger(ctx, LOG_INFO, sigmsg, "SIGUSR1", "renewing");
+		dhcpcd_renew(ctx);
 		return;
 	case SIGUSR2:
 		logger_close(ctx);
@@ -1220,7 +1285,7 @@ signal_cb(int sig, void *arg)
 	}
 
 	if (!(ctx->options & DHCPCD_TEST))
-		stop_all_interfaces(ctx, do_release);
+		stop_all_interfaces(ctx, opts);
 	eloop_exit(ctx->eloop, exit_code);
 }
 #endif
@@ -1234,6 +1299,8 @@ dhcpcd_getinterfaces(void *arg)
 
 	len = 0;
 	TAILQ_FOREACH(ifp, fd->ctx->ifaces, next) {
+		if (!ifp->active)
+			continue;
 		len++;
 		if (D_STATE_RUNNING(ifp))
 			len++;
@@ -1248,6 +1315,8 @@ dhcpcd_getinterfaces(void *arg)
 		return;
 	eloop_event_remove_writecb(fd->ctx->eloop, fd->fd);
 	TAILQ_FOREACH(ifp, fd->ctx->ifaces, next) {
+		if (!ifp->active)
+			continue;
 		if (send_interface(fd, ifp) == -1)
 			logger(ifp->ctx, LOG_ERR,
 			    "send_interface %d: %m", fd->fd);
@@ -1259,8 +1328,8 @@ dhcpcd_handleargs(struct dhcpcd_ctx *ctx, struct fd_list *fd,
     int argc, char **argv)
 {
 	struct interface *ifp;
-	int do_exit = 0, do_release = 0, do_reboot = 0;
-	int opt, oi = 0;
+	unsigned long long opts;
+	int opt, oi, do_reboot, do_renew;
 	size_t len, l;
 	char *tmp, *p;
 
@@ -1309,6 +1378,9 @@ dhcpcd_handleargs(struct dhcpcd_ctx *ctx, struct fd_list *fd,
 	free(tmp);
 
 	optind = 0;
+	oi = 0;
+	opts = 0;
+	do_reboot = do_renew = 0;
 	while ((opt = getopt_long(argc, argv, IF_OPTS, cf_options, &oi)) != -1)
 	{
 		switch (opt) {
@@ -1316,36 +1388,53 @@ dhcpcd_handleargs(struct dhcpcd_ctx *ctx, struct fd_list *fd,
 			/* Assumed if below not set */
 			break;
 		case 'k':
-			do_release = 1;
+			opts |= DHCPCD_RELEASE;
 			break;
 		case 'n':
 			do_reboot = 1;
 			break;
+		case 'p':
+			opts |= DHCPCD_PERSISTENT;
+			break;
 		case 'x':
-			do_exit = 1;
+			opts |= DHCPCD_EXITING;
+			break;
+		case 'N':
+			do_renew = 1;
 			break;
 		}
 	}
 
-	if (do_release || do_exit) {
+	if (opts & (DHCPCD_EXITING | DHCPCD_RELEASE)) {
 		if (optind == argc) {
-			stop_all_interfaces(ctx, do_release);
+			stop_all_interfaces(ctx, opts);
 			eloop_exit(ctx->eloop, EXIT_SUCCESS);
 			return 0;
 		}
 		for (oi = optind; oi < argc; oi++) {
 			if ((ifp = if_find(ctx->ifaces, argv[oi])) == NULL)
 				continue;
-			if (do_release) {
-				ifp->options->options |= DHCPCD_RELEASE;
+			ifp->options->options |= opts;
+			if (opts & DHCPCD_RELEASE)
 				ifp->options->options &= ~DHCPCD_PERSISTENT;
-			}
-			ifp->options->options |= DHCPCD_EXITING;
 			stop_interface(ifp);
 		}
 		return 0;
 	}
 
+	if (do_renew) {
+		if (optind == argc) {
+			dhcpcd_renew(ctx);
+			return 0;
+		}
+		for (oi = optind; oi < argc; oi++) {
+			if ((ifp = if_find(ctx->ifaces, argv[oi])) == NULL)
+				continue;
+			dhcpcd_ifrenew(ifp);
+		}
+		return 0;
+	}
+
 	reload_config(ctx);
 	/* XXX: Respect initial commandline options? */
 	reconf_reboot(ctx, do_reboot, argc, argv, optind - 1);
@@ -1407,7 +1496,9 @@ main(int argc, char **argv)
 	ctx.udp_fd = -1;
 #endif
 	i = 0;
-	while ((opt = getopt_long(argc, argv, IF_OPTS, cf_options, &oi)) != -1)
+	while ((opt = getopt_long(argc, argv,
+	    ctx.options & DHCPCD_PRINT_PIDFILE ? NOERR_IF_OPTS : IF_OPTS,
+	    cf_options, &oi)) != -1)
 	{
 		switch (opt) {
 		case '4':
@@ -1419,16 +1510,12 @@ main(int argc, char **argv)
 		case 'f':
 			ctx.cffile = optarg;
 			break;
-#ifdef USE_SIGNALS
-		case 'g':
-			sig = SIGUSR1;
-			siga = "USR1";
-			break;
 		case 'j':
 			ctx.logfile = strdup(optarg);
 			logger_close(&ctx);
 			logger_open(&ctx);
 			break;
+#ifdef USE_SIGNALS
 		case 'k':
 			sig = SIGALRM;
 			siga = "ARLM";
@@ -1437,11 +1524,24 @@ main(int argc, char **argv)
 			sig = SIGHUP;
 			siga = "HUP";
 			break;
+		case 'g':
+		case 'p':
+			/* Force going via command socket as we're
+			 * out of user definable signals. */
+			i = 4;
+			break;
 		case 'x':
 			sig = SIGTERM;
-			siga = "TERM";;
+			siga = "TERM";
+			break;
+		case 'N':
+			sig = SIGUSR1;
+			siga = "USR1";
 			break;
 #endif
+		case 'P':
+			ctx.options |= DHCPCD_PRINT_PIDFILE;
+			break;
 		case 'T':
 			i = 1;
 			break;
@@ -1452,6 +1552,8 @@ main(int argc, char **argv)
 			i = 2;
 			break;
 		case '?':
+			if (ctx.options & DHCPCD_PRINT_PIDFILE)
+				continue;
 			usage();
 			goto exit_failure;
 		}
@@ -1463,10 +1565,15 @@ main(int argc, char **argv)
 	ctx.ifv = argv + optind;
 
 	ifo = read_config(&ctx, NULL, NULL, NULL);
-	if (ifo == NULL)
+	if (ifo == NULL) {
+		if (ctx.options & DHCPCD_PRINT_PIDFILE)
+			goto printpidfile;
 		goto exit_failure;
+	}
 	opt = add_options(&ctx, NULL, ifo, argc, argv);
 	if (opt != 1) {
+		if (ctx.options & DHCPCD_PRINT_PIDFILE)
+			goto printpidfile;
 		if (opt == 0)
 			usage();
 		goto exit_failure;
@@ -1500,8 +1607,8 @@ main(int argc, char **argv)
 #endif
 		goto exit_success;
 	}
-	ctx.options = ifo->options;
-	if (i != 0) {
+	ctx.options |= ifo->options;
+	if (i == 1 || i == 3) {
 		if (i == 1)
 			ctx.options |= DHCPCD_TEST;
 		else
@@ -1516,17 +1623,9 @@ main(int argc, char **argv)
 
 	if (ctx.options & DHCPCD_DEBUG)
 		logger_mask(&ctx, LOG_UPTO(LOG_DEBUG));
-	if (ctx.options & DHCPCD_QUIET) {
-		i = open(_PATH_DEVNULL, O_RDWR);
-		if (i == -1)
-			logger(&ctx, LOG_ERR, "%s: open: %m", __func__);
-		else {
-			dup2(i, STDERR_FILENO);
-			close(i);
-		}
-	}
 
 	if (!(ctx.options & (DHCPCD_TEST | DHCPCD_DUMPLEASE))) {
+printpidfile:
 		/* If we have any other args, we should run as a single dhcpcd
 		 *  instance for that interface. */
 		if (optind == argc - 1 && !(ctx.options & DHCPCD_MASTER)) {
@@ -1556,6 +1655,10 @@ main(int argc, char **argv)
 			    PIDFILE, "", "", "");
 			ctx.options |= DHCPCD_MASTER;
 		}
+		if (ctx.options & DHCPCD_PRINT_PIDFILE) {
+			printf("%s\n", ctx.pidfile);
+			goto exit_success;
+		}
 	}
 
 	if (chdir("/") == -1)
@@ -1568,28 +1671,37 @@ main(int argc, char **argv)
 		goto exit_failure;
 	}
 
+	/* Open our persistent sockets.
+	 * This is needed early for dumping leases on valid interfaces. */
+	if (if_opensockets(&ctx) == -1) {
+		logger(&ctx, LOG_ERR, "if_opensockets: %m");
+		goto exit_failure;
+	}
+
 	if (ctx.options & DHCPCD_DUMPLEASE) {
-		if (optind != argc - 1) {
-			logger(&ctx, LOG_ERR,
-			    "dumplease requires an interface");
-			goto exit_failure;
+		if (optind != argc) {
+			/* We need to try and find the interface so we can load
+			 * the hardware address to compare automated IAID */
+			ctx.ifaces = if_discover(&ctx,
+			    argc - optind, argv + optind);
+		} else {
+			if ((ctx.ifaces = malloc(sizeof(*ctx.ifaces))) != NULL)
+				TAILQ_INIT(ctx.ifaces);
 		}
-		i = 0;
-		/* We need to try and find the interface so we can
-		 * load the hardware address to compare automated IAID */
-		ctx.ifaces = if_discover(&ctx, 1, argv + optind);
 		if (ctx.ifaces == NULL) {
 			logger(&ctx, LOG_ERR, "if_discover: %m");
 			goto exit_failure;
 		}
-		ifp = TAILQ_FIRST(ctx.ifaces);
+		ifp = if_find(ctx.ifaces, argv[optind]);
 		if (ifp == NULL) {
 			ifp = calloc(1, sizeof(*ifp));
 			if (ifp == NULL) {
 				logger(&ctx, LOG_ERR, "%s: %m", __func__);
 				goto exit_failure;
 			}
-			strlcpy(ctx.pidfile, argv[optind], sizeof(ctx.pidfile));
+			if (optind != argc)
+				strlcpy(ctx.pidfile, argv[optind],
+				    sizeof(ctx.pidfile));
 			ifp->ctx = &ctx;
 			TAILQ_INSERT_HEAD(ctx.ifaces, ifp, next);
 			if (family == 0) {
@@ -1600,13 +1712,14 @@ main(int argc, char **argv)
 			}
 		}
 		configure_interface(ifp, ctx.argc, ctx.argv, 0);
+		i = 0;
 		if (family == 0 || family == AF_INET) {
 			if (dhcp_dump(ifp) == -1)
-				i = 1;
+				i = -1;
 		}
 		if (family == 0 || family == AF_INET6) {
 			if (dhcp6_dump(ifp) == -1)
-				i = 1;
+				i = -1;
 		}
 		if (i == -1)
 			goto exit_failure;
@@ -1614,8 +1727,11 @@ main(int argc, char **argv)
 	}
 
 #ifdef USE_SIGNALS
-	if (!(ctx.options & DHCPCD_TEST) &&
-	    (sig == 0 || ctx.ifc != 0))
+	/* Test against siga instead of sig to avoid gcc
+	 * warning about a bogus potential signed overflow.
+	 * The end result will be the same. */
+	if ((siga == NULL || i == 4 || ctx.ifc != 0) &&
+	    !(ctx.options & DHCPCD_TEST))
 	{
 #endif
 		if (ctx.options & DHCPCD_MASTER)
@@ -1652,14 +1768,14 @@ main(int argc, char **argv)
 			logger(&ctx, LOG_INFO, "sending signal %s to pid %d",
 			    siga, pid);
 		if (pid == 0 || kill(pid, sig) != 0) {
-			if (sig != SIGHUP && errno != EPERM)
+			if (sig != SIGHUP && sig != SIGUSR1 && errno != EPERM)
 				logger(&ctx, LOG_ERR, ""PACKAGE" not running");
 			if (pid != 0 && errno != ESRCH) {
 				logger(&ctx, LOG_ERR, "kill: %m");
 				goto exit_failure;
 			}
 			unlink(ctx.pidfile);
-			if (sig != SIGHUP)
+			if (sig != SIGHUP && sig != SIGUSR1)
 				goto exit_failure;
 		} else {
 			struct timespec ts;
@@ -1709,9 +1825,11 @@ main(int argc, char **argv)
 			/* Lock the file so that only one instance of dhcpcd
 			 * runs on an interface */
 			if (flock(ctx.pid_fd, LOCK_EX | LOCK_NB) == -1) {
-				logger(&ctx, LOG_ERR, "flock `%s': %m", ctx.pidfile);
-				close(ctx.pid_fd);
-				ctx.pid_fd = -1;
+				logger(&ctx, LOG_ERR, "flock `%s': %m",
+				    ctx.pidfile);
+				/* We don't want to unlink the pidfile as
+				 * another dhcpcd instance could be using it. */
+				ctx.pidfile[0] = '\0';
 				goto exit_failure;
 			}
 #endif
@@ -1720,8 +1838,9 @@ main(int argc, char **argv)
 			    fcntl(ctx.pid_fd, F_SETFD, opt | FD_CLOEXEC) == -1)
 			{
 				logger(&ctx, LOG_ERR, "fcntl: %m");
-				close(ctx.pid_fd);
-				ctx.pid_fd = -1;
+				/* We don't want to unlink the pidfile as
+				 * another dhcpcd instance could be using it. */
+				ctx.pidfile[0] = '\0';
 				goto exit_failure;
 			}
 #endif
@@ -1749,7 +1868,7 @@ main(int argc, char **argv)
 	    dhcpcd_signals, dhcpcd_signals_len,
 	    signal_cb, &ctx) == -1)
 	{
-		logger(&ctx, LOG_ERR, "eloop_signal_mask: %m");
+		logger(&ctx, LOG_ERR, "eloop_signal_set_cb: %m");
 		goto exit_failure;
 	}
 	/* Save signal mask, block and redirect signals to our handler */
@@ -1764,11 +1883,8 @@ main(int argc, char **argv)
 	if (ctx.ifc == 1 && !(ctx.options & DHCPCD_BACKGROUND))
 		ctx.options |= DHCPCD_WAITIP;
 
-	/* Open our persistent sockets. */
-	if (if_opensockets(&ctx) == -1) {
-		logger(&ctx, LOG_ERR, "if_opensockets: %m");
-		goto exit_failure;
-	}
+	/* Start handling kernel messages for interfaces, addreses and
+	 * routes. */
 	eloop_event_add(ctx.eloop, ctx.link_fd, handle_link, &ctx, NULL, NULL);
 
 	/* Start any dev listening plugin which may want to
@@ -1783,12 +1899,17 @@ main(int argc, char **argv)
 		goto exit_failure;
 	}
 	for (i = 0; i < ctx.ifc; i++) {
-		if (if_find(ctx.ifaces, ctx.ifv[i]) == NULL)
+		if ((ifp = if_find(ctx.ifaces, ctx.ifv[i])) == NULL ||
+		    !ifp->active)
 			logger(&ctx, LOG_ERR,
 			    "%s: interface not found or invalid",
 			    ctx.ifv[i]);
 	}
-	if (TAILQ_FIRST(ctx.ifaces) == NULL) {
+	TAILQ_FOREACH(ifp, ctx.ifaces, next) {
+		if (ifp->active)
+			break;
+	}
+	if (ifp == NULL) {
 		if (ctx.ifc == 0)
 			logger(&ctx, LOG_ERR, "no valid interfaces found");
 		else
@@ -1801,7 +1922,8 @@ main(int argc, char **argv)
 	}
 
 	TAILQ_FOREACH(ifp, ctx.ifaces, next) {
-		dhcpcd_initstate1(ifp, argc, argv, 0);
+		if (ifp->active)
+			dhcpcd_initstate1(ifp, argc, argv, 0);
 	}
 
 	if (ctx.options & DHCPCD_BACKGROUND && dhcpcd_daemonise(&ctx))
@@ -1809,18 +1931,26 @@ main(int argc, char **argv)
 
 	opt = 0;
 	TAILQ_FOREACH(ifp, ctx.ifaces, next) {
-		run_preinit(ifp);
-		if (ifp->carrier != LINK_DOWN)
-			opt = 1;
+		if (ifp->active) {
+			run_preinit(ifp);
+			if (!(ifp->options->options & DHCPCD_LINK) ||
+			    ifp->carrier != LINK_DOWN)
+				opt = 1;
+		}
 	}
 
 	if (!(ctx.options & DHCPCD_BACKGROUND)) {
 		if (ctx.options & DHCPCD_MASTER)
 			t = ifo->timeout;
-		else if ((ifp = TAILQ_FIRST(ctx.ifaces)))
-			t = ifp->options->timeout;
-		else
+		else {
 			t = 0;
+			TAILQ_FOREACH(ifp, ctx.ifaces, next) {
+				if (ifp->active) {
+					t = ifp->options->timeout;
+					break;
+				}
+			}
+		}
 		if (opt == 0 &&
 		    ctx.options & DHCPCD_LINK &&
 		    !(ctx.options & DHCPCD_WAITIP))
@@ -1842,8 +1972,9 @@ main(int argc, char **argv)
 
 	if_sortinterfaces(&ctx);
 	TAILQ_FOREACH(ifp, ctx.ifaces, next) {
-		eloop_timeout_add_sec(ctx.eloop, 0,
-		    dhcpcd_prestartinterface, ifp);
+		if (ifp->active)
+			eloop_timeout_add_sec(ctx.eloop, 0,
+			    dhcpcd_prestartinterface, ifp);
 	}
 
 	i = eloop_start(ctx.eloop, &ctx.sigset);
@@ -1885,7 +2016,6 @@ exit1:
 		close(ctx.pf_link_fd);
 #endif
 
-
 	free_options(ifo);
 	free_globals(&ctx);
 	ipv4_ctxfree(&ctx);
@@ -1895,7 +2025,8 @@ exit1:
 		logger(&ctx, LOG_ERR, "control_stop: %m:");
 	if (ctx.pid_fd != -1) {
 		close(ctx.pid_fd);
-		unlink(ctx.pidfile);
+		if (ctx.pidfile[0] != '\0')
+			unlink(ctx.pidfile);
 	}
 	eloop_free(ctx.eloop);
 
